#include "omp.h"
#include "stdio.h"
#include "stdlib.h"
#include "math.h"
#include <sys/time.h>

float *d_u0, *d_u1, *d_u2;
#define MIN(a,b) ((a)<=(b) ? (a) : (b))
#define MAX(a,b) ((a)>=(b) ? (a) : (b))
#define CEIL_DIV(a,d) (((a)+(d)-1)/(d)) // modular arithmetic trick to ceiling divide 

#define D_U0(y,x) d_u0[(y)*(width)+(x)]
#define D_U1(y,x) d_u1[(y)*(width)+(x)]
#define D_U2(y,x) d_u2[(y)*(width)+(x)]

float time_diff_in_ms(struct timeval t1, struct timeval t2) {
    float diff_sec = t2.tv_sec - t1.tv_sec;
    float diff_usec = t2.tv_usec - t1.tv_usec;
    return (diff_sec * 1000.0f) + (diff_usec / 1000.0f);
}

// point stencil for wave equation, only on the inner points
void inner_kernel(float c, float *d_u0, float *d_u1, float *d_u2, int width, int height){
    #pragma omp parallel for
    for (int y = 1; y < height - 1; y++){
        for (int x = 1; x < width - 1; x++){
            float mid1 = D_U1(y, x);
            float east1  = D_U1(y, x + 1);
            float west1  = D_U1(y, x - 1);
            float north1 = D_U1(y - 1, x);
            float south1 = D_U1(y + 1, x);

            D_U2(y, x) = 2.0f * mid1 - D_U0(y, x) + (c * c) * (-4.0f * mid1 + east1 + south1 + north1 + west1);
        }
    }
}

void initialize_state(float *d_u0, float *d_u1, float *d_u2, int width, int height){
    #pragma omp parallel for
    for (int y = 0; y < height; y++){
        for (int x = 0; x < width; x++){
            D_U0(y, x) = -exp(
                -pow(x - ((float)width / 2.0), 2.0) / (float)width 
                -pow(y - ((float)height / 2.0), 2.0) / (float)height);
            D_U1(y, x) = D_U0(y, x);
            D_U2(y, x) = D_U0(y, x);
        }
    }
}
int main(){
    setbuf(stdout, NULL);
    int sidelength = 16384;
    int start = 25;
    int end =  16384;
    const int runs = 100;
    const int increment = 25;

    d_u0 = (float*)malloc(sidelength*sidelength*sizeof(float));
    d_u1 = (float*)malloc(sidelength*sidelength*sizeof(float));
    d_u2 = (float*)malloc(sidelength*sidelength*sizeof(float));

    initialize_state(d_u0, d_u1, d_u2, sidelength, sidelength);
    struct timeval t1, t2;

    gettimeofday(&t1, NULL);

    for (int w = start; w <= end; w+=increment){
        gettimeofday(&t1, NULL);
        for (int r = 0; r < runs; r++){
            inner_kernel(2.0, d_u0, d_u1, d_u2, w, w);

            float *tmp = d_u0;
            d_u0 = d_u1;
            d_u1 = d_u2;
            d_u2 = tmp;
        }
        gettimeofday(&t2, NULL);
        printf("%d: %f\n",w, time_diff_in_ms(t1, t2)/runs);
    }

    // fclose(f);
    free(d_u0);
    free(d_u1);
    free(d_u2);
}