#include <cuda.h>
#include <cuda_runtime.h>

#include <iostream>
#include <fstream>
#include <unistd.h>

cudaEvent_t start, stop;
cudaStream_t stream;
dim3 fullblock, lineblock;
dim3 fullgrid, lingrid, cartgrid; 
#include <sys/time.h>

#define gpuErrchk(ans) { gpuAssert((ans), __FILE__, __LINE__); }
inline void gpuAssert(cudaError_t code, const char *file, int line, bool abort=true)
{
   if (code != cudaSuccess) 
   {
      fprintf(stderr,"GPUassert: %s %s %d\n", cudaGetErrorString(code), file, line);
      if (abort) exit(code);
   }
}

float time_diff_in_ms(struct timeval t1, struct timeval t2) {
    float diff_sec = t2.tv_sec - t1.tv_sec;
    float diff_usec = t2.tv_usec - t1.tv_usec;
    return (diff_sec * 1000.0f) + (diff_usec / 1000.0f);
}


float *d_u0, *d_u1, *d_u2;
#define MIN(a,b) ((a)<=(b) ? (a) : (b))
#define MAX(a,b) ((a)>=(b) ? (a) : (b))
#define CEIL_DIV(a,d) (((a)+(d)-1)/(d)) // modular arithmetic trick to ceiling divide 

#define D_U0(y,x) d_u0[((y)+1)*(2+width)+(x)+1]
#define D_U1(y,x) d_u1[((y)+1)*(2+width)+(x)+1]
#define D_U2(y,x) d_u2[((y)+1)*(2+width)+(x)+1]

__global__ void inner_kernel(float c, float *d_u0, float *d_u1, float *d_u2, int width, int height){
    int x = threadIdx.x + blockDim.x * blockIdx.x;
    int y = threadIdx.y + blockDim.y * blockIdx.y + 1;

    if (x < width && y < height - 1){
        float mid1 = D_U1(y, x);
        // logic to fetch value from neighbor or mirror on the wall
        float east1  = (x == width - 1  ? D_U1(y, x - 1) : D_U1(y, x + 1));
        float west1  = (x == 0          ? D_U1(y, x + 1) : D_U1(y, x - 1));
        float north1 = D_U1(y - 1, x);
        float south1 = D_U1(y + 1, x);

        // handle the standard inner case with 5 point star stencil
        D_U2(y, x) = 2.0f * mid1 - D_U0(y, x) + (c * c) * (-4.0f * mid1 + east1 + south1 + north1 + west1);
    }
}

// __global__ void linear_outer(float c, float *d_u0, float *d_u1, float *d_u2, int width, int gpuheight){
//     int thr = threadIdx.x + blockDim.x * blockIdx.x;
//     if (thr < 2 * width){
//         int x, y;
//         if (thr < width){ // northern border
//             y = 0;
//             x = thr;
//         }
//         else if (thr < 2 * width){ // southern border
//             y = gpuheight - 1;
//             x = thr - width;
//         }
//         float mid1   = D_U1(y, x);
//         float east1  = (x == width - 1         ? D_U1(y, x - 1) : D_U1(y, x + 1));
//         float west1  = (x == 0                 ? D_U1(y, x + 1) : D_U1(y, x - 1));
//         float north1 = (y == 0                 ? D_U1(y + 1, x) : D_U1(y - 1, x));
//         float south1 = (y == gpuheight - 1     ? D_U1(y - 1, x) : D_U1(y + 1, x));

//         D_U2(y, x) = 2.0f * mid1 - D_U0(y, x) + (c * c) * (-4.0f * mid1 + east1 + south1 + north1 + west1);
//     }
// }

// __global__ void cart_outer(float c, float *d_u0, float *d_u1, float *d_u2, int width, int height){
//     int thr = threadIdx.x + blockDim.x * blockIdx.x;
//     if (thr < 2 * width + 2 * height - 4){
//         int x, y;
//         if (thr < width){ // northern border
//             y = 0;
//             x = thr;
//         }
//         else if (thr < 2 * width){ // southern border
//             y = height - 1;
//             x = thr - width;
//         }
//         else if (thr < 2 * width + height - 2){ // western border
//             x = 0;
//             y = thr - 2 * width + 1;
//         }
//         else if (thr < 2 * width + 2 * height - 4){ // eastern border
//             x = width - 1;
//             y = thr - 2 * width - (height - 2) + 1;
//         }
//         float mid1   = D_U1(y, x);
//         float east1  = (x == width - 1         ? D_U1(y, x - 1) : D_U1(y, x + 1));
//         float west1  = (x == 0                 ? D_U1(y, x + 1) : D_U1(y, x - 1));
//         float north1 = (y == 0                 ? D_U1(y + 1, x) : D_U1(y - 1, x));
//         float south1 = (y == height - 1 ? D_U1(y - 1, x) : D_U1(y + 1, x));

//         D_U2(y, x) = 2.0f * mid1 - D_U0(y, x) + (c * c) * (-4.0f * mid1 + east1 + south1 + north1 + west1);
//     }
// }

__global__ void initialize_state(float *d_u0, float *d_u1, float *d_u2, int width){
    int x = threadIdx.x + blockDim.x * blockIdx.x;
    int y = threadIdx.y + blockDim.y * blockIdx.y;
    if (x <= width && y <= width){
        D_U0(y, x) = -exp(
            -pow(x - ((float)width / 2.0), 2.0) / (float)width 
            -pow(y - ((float)width / 2.0), 2.0) / (float)width);
        D_U1(y, x) = D_U0(y, x);
        D_U2(y, x) = D_U0(y, x);
    }
}

__host__ void cuda_terminate(){
    cudaFree(d_u0);
    cudaFree(d_u1);
    cudaFree(d_u2);
    cudaEventDestroy(start);
    cudaEventDestroy(stop);
    cudaStreamDestroy(stream);
}

int main(){
    cudaEventCreate(&start);
    cudaEventCreate(&stop);
    cudaStreamCreate(&stream);

    int sidelength = 1575;//16384;
    const int runs = 25;//100;
    const int increment = 1000;//25
    float run_time;

    fullblock = dim3(32, 32);
    // lineblock = dim3(1024);
    fullgrid = dim3(CEIL_DIV(sidelength, 32), CEIL_DIV(sidelength, 32));

    if (0 != cudaMalloc((void **)&d_u0, (sidelength+2)*(sidelength+2)*sizeof(float))){
        printf("malloc failed!\n");
        exit(1);
    }
    if (0 != cudaMalloc((void **)&d_u1, (sidelength+2)*(sidelength+2)*sizeof(float))){
        printf("malloc failed!\n");
        exit(1);
    }
    if (0 != cudaMalloc((void **)&d_u2, (sidelength+2)*(sidelength+2)*sizeof(float))){
        printf("malloc failed!\n");
        exit(1);
    }

    initialize_state<<<fullgrid, fullblock, 0, stream>>>(d_u0, d_u1, d_u2, sidelength);
    timeval a, b;
    for (int w = 575; w <= sidelength; w+=increment){
        fullgrid = dim3(CEIL_DIV(w,32),CEIL_DIV(w,32));
        // lingrid = dim3(CEIL_DIV(2*w,1024));
        // cartgrid = dim3(CEIL_DIV(4*w,1024));
        gettimeofday(&a, NULL);
        inner_kernel<<<fullgrid, fullblock, 0, stream>>>(2.0, d_u0, d_u1, d_u2, sidelength, sidelength); // warmup run
        gpuErrchk (cudaPeekAtLastError());
        // linear_outer<<<lingrid, lineblock, 0, stream>>>(2.0, d_u0, d_u1, d_u2, sidelength, sidelength);
        // cart_outer<<<cartgrid, lineblock, 0, stream>>>(2.0, d_u0, d_u1, d_u2, sidelength, sidelength);

        cudaEventRecord(start, stream);
        for (int r = 0; r < runs; r++){
            inner_kernel<<<fullgrid, fullblock, 0, stream>>>(2.0, d_u0, d_u1, d_u2, sidelength, sidelength);
        }
        cudaEventRecord(stop, stream);
        cudaEventSynchronize(stop);
        gettimeofday(&b, NULL);
        cudaEventElapsedTime(&run_time, start, stop);
        gpuErrchk (cudaPeekAtLastError());
        std::cout << "INNER: " << run_time <<  " " << time_diff_in_ms(a, b) <<  std::endl;
    }

    cuda_terminate();
}