#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <omp.h>
#include <sched.h>

long N = 1e9;

struct timeval times[8];

void COPY(float* __restrict__ src, float* __restrict__ dst){
    gettimeofday(&times[0], NULL);
    #ifdef _OPENMP
    #pragma omp parallel for
    #endif
    for (int i = 0; i < N; i++){
        dst[i] = src[i];
    }
    gettimeofday(&times[1], NULL);
}

// dst = c*a
void SCALE(float* __restrict__ a, float c, float* __restrict__ dst){
    gettimeofday(&times[2], NULL);
    #ifdef _OPENMP
    #pragma omp parallel for
    #endif
    for (int i = 0; i < N; i++){
        dst[i] = c*a[i];
    }
    gettimeofday(&times[3], NULL);
}

// dst = a + b
void ADD(float* __restrict__ a, float* __restrict__ b, float* __restrict__ dst){
    gettimeofday(&times[4], NULL);
    #ifdef _OPENMP
    #pragma omp parallel for
    #endif
    for (int i = 0; i < N; i++){
        dst[i] = a[i] + b[i];
    }
    gettimeofday(&times[5], NULL);
}

// dst = a + c*b
void TRIAD(float* __restrict__ a, float* __restrict__ b, float c, float* __restrict__ dst){
    gettimeofday(&times[6], NULL);
    #ifdef _OPENMP
    #pragma omp parallel for
    #endif
    for (int i = 0; i < N; i++){
        dst[i] = a[i] + c*b[i];
    }
    gettimeofday(&times[7], NULL);
}

double diff(struct timeval x , struct timeval y)
{
    double x_ms , y_ms , diff;
      
    x_ms = (double)x.tv_sec*1000000 + (double)x.tv_usec;
    y_ms = (double)y.tv_sec*1000000 + (double)y.tv_usec;
      
    diff = (double)y_ms - (double)x_ms;
      
    return diff;
}

int main(){
    FILE *fp;
    char path[1035];

    // print cpu info
    fp = popen("lscpu", "r");
    if (fp == NULL) {
        printf("Failed to run command\n");
        exit(1);
    }
    while (fgets(path, sizeof(path)-1, fp) != NULL) {
        printf("%s", path);
    }
    pclose(fp);

    const char* ops[4] = {"COPY", "SCALE", "ADD", "TRIAD"};
    double num_of_float_transfers[4] = {2.0, 2.0, 3.0, 3.0};

    float *a = (float*)malloc(sizeof(float)*N), 
          *b = (float*)malloc(sizeof(float)*N),
          *c = (float*)malloc(sizeof(float)*N),
          scale=3.f;
    #ifdef _OPENMP
    #pragma omp parallel for
    #endif
    for (int i = 0; i < N; i++){
        a[i] = (i%2==0 ? 2.0f : 5.3f);
        b[i] = (i%2==0 ? 3.0f : 4.4f);
        c[i] = (i%2==0 ? 4.0f : 3.5f);
    }

    // show core and thread mapping
    int k = 0;
    #pragma omp parallel
    #pragma omp atomic 
        k++;
    printf("number of threads %d\n", k);

    int threadtocore[100];
    #pragma omp parallel
    {
        int tid = omp_get_thread_num();
        // unsigned int core, numanode;
        int core = sched_getcpu();
        if (core != -1)
            threadtocore[tid]=core;
        // if ( 0 == getcpu ( &core, &numanode ) )
        else
            printf ( "Unable to determine core #\n" );
    }
    for (int i = 0; i < k; i++){
        printf("thr %d on core %u\n", i, threadtocore[i]);
    }

    float best[4] = {1000000000, 1000000000, 1000000000, 1000000000};

    for (int i = 0; i < 100; i++){
        printf("IT: %d\n", i);
        COPY(a, b);
        SCALE(a, scale, b);
        ADD(a, b, c);
        TRIAD(a, b, scale, c);

        for (int j = 0; j < 4; j++){
            double sec = diff(times[2*j], times[2*j+1])/1e6;
            double GB = ((double) N * sizeof(float) * (double) num_of_float_transfers[j])/(double)1e9;
            double GB_per_sec = GB/sec;
            // printf("%s: %f\n", ops[j], GB_per_sec);
            best[j] = (best[j] < sec ? best[j] : sec);
        }
    }

    for (int i = 0; i < 4; i++){
        double sec = best[i];
        double GB = ((double) N * sizeof(float) * (double) num_of_float_transfers[i])/(double)1e9;
        double GB_per_sec = GB/sec;

        printf("%s: %13lfGBs, Mem transfer: %lfGB, Runtime: %lfs\n", ops[i], GB_per_sec, GB, sec);
    }
}