#include <cuda.h>
#include <cuda_runtime.h>

extern "C" {
    #include "heterogeneous_FDM_config.h"
}
cudaStream_t comp_stream, mem_stream;
cudaEvent_t start, stop, border_done, packing_done;
dim3 grid, block;
dim3 full_border_grid, single_border_grid, full_border_block, single_border_block;
dim3 inner_grid;

// kernel computing the entire grid
__global__ void full_kernel(real_t c, real_t *d_u0, real_t *d_u1, real_t *d_u2, int_t width, int_t height, char ns){
    int_t x = threadIdx.x + blockDim.x * blockIdx.x;
    int_t y = threadIdx.y + blockDim.y * blockIdx.y;

    if (x < width && y < height){
        real_t mid1 = D_U1(y, x);
        // logic to fetch value from neighbor or mirror on the wall
        real_t east1  = (             x == width - 1  ? D_U1(y, x - 1) : D_U1(y, x + 1));
        real_t west1  = (             x == 0          ? D_U1(y, x + 1) : D_U1(y, x - 1));
        real_t north1 = (ON_N_EDGE && y == 0          ? D_U1(y + 1, x) : D_U1(y - 1, x));
        real_t south1 = (ON_S_EDGE && y == height - 1 ? D_U1(y - 1, x) : D_U1(y + 1, x));

        // handle the standard inner case with 5 point star stencil
        D_U2(y, x) = 2.0f * mid1 - D_U0(y, x) + (c * c) * (-4.0f * mid1 + east1 + south1 + north1 + west1);
    }
}

// kernel computing the inner points
__global__ void inner_kernel(real_t c, real_t *d_u0, real_t *d_u1, real_t *d_u2, int_t width, int_t height){
    // offset by one to make sure our index is not on western or nothern boundary
    int_t x = threadIdx.x + blockDim.x * blockIdx.x;
    int_t y = 1 + threadIdx.y + blockDim.y * blockIdx.y;

    // only make computations on inner points
    if (x < width && y < height - 1){
        real_t mid1 = D_U1(y, x);
        real_t east1  = (x == width - 1  ? D_U1(y, x - 1) : D_U1(y, x + 1));
        real_t west1  = (x == 0          ? D_U1(y, x + 1) : D_U1(y, x - 1));
        real_t north1 = D_U1(y - 1, x);
        real_t south1 = D_U1(y + 1, x);

        // handle the standard inner case with 5 point star stencil
        D_U2(y, x) = 2.0f * mid1 - D_U0(y, x) + (c * c) * (-4.0f * mid1 + east1 + south1 + north1 + west1);
    }
}

// kernel computing the values on the border
__global__ void outer_kernel(real_t c, real_t *d_u0, real_t *d_u1, real_t *d_u2, int_t width, int_t gpuheight, int_t cpugpuheight, char ns){
    int_t thr = threadIdx.x + blockDim.x * blockIdx.x;
    if (thr < 2 * width){
        int_t x, y;
        if (thr < width){ // northern border
            y = 0;
            x = thr;
        }
        else if (thr < 2 * width){ // southern border
            y = gpuheight - 1;
            x = thr - width;
        }
        real_t mid1   = D_U1(y, x);
        real_t east1  = (             x == width - 1         ? D_U1(y, x - 1) : D_U1(y, x + 1));
        real_t west1  = (             x == 0                 ? D_U1(y, x + 1) : D_U1(y, x - 1));
        real_t north1 = (ON_N_EDGE && y == 0                 ? D_U1(y + 1, x) : D_U1(y - 1, x));
        real_t south1 = (ON_S_EDGE && y == cpugpuheight - 1        ? D_U1(y - 1, x) : D_U1(y + 1, x));

        D_U2(y, x) = 2.0f * mid1 - D_U0(y, x) + (c * c) * (-4.0f * mid1 + east1 + south1 + north1 + west1);
    }
}

//kernel computing the entire grid, but not the intra border (inteded for single node, GPU + CPU)
__global__ void gpu_share_without_intra_border_kernel(real_t c, real_t *d_u0, real_t *d_u1, real_t *d_u2, int_t width, int_t gpu_height, int_t height){
    int_t x = threadIdx.x + blockDim.x * blockIdx.x;
    int_t y = threadIdx.y + blockDim.y * blockIdx.y;

    // off by one error her, du må ha en variant av hver funksjon, kan kanskje fikses med preprossesering
    if (x < width && y < gpu_height - 1){
        real_t mid1 = D_U1(y, x);
        // logic to fetch value from neighbor or mirror on the wall
        real_t east1  = (x == width - 1  ? D_U1(y, x - 1) : D_U1(y, x + 1));
        real_t west1  = (x == 0          ? D_U1(y, x + 1) : D_U1(y, x - 1));
        real_t north1 = (y == 0          ? D_U1(y + 1, x) : D_U1(y - 1, x));
        real_t south1 = (y == height - 1 ? D_U1(y - 1, x) : D_U1(y + 1, x));

        // handle the standard inner case with 5 point star stencil
        D_U2(y, x) = 2.0f * mid1 - D_U0(y, x) + (c * c) * (-4.0f * mid1 + east1 + south1 + north1 + west1);
    }
}

__global__ void intra_border_kernel(real_t c,real_t *d_u0,real_t *d_u1,real_t *d_u2,int_t width,int_t gpu_height){
    int_t x = threadIdx.x + blockDim.x * blockIdx.x;
    int_t y = gpu_height - 1;
    if (x < width){
        real_t mid1 = D_U1(y, x);
        // logic to fetch value from neighbor or mirror on the wall
        real_t east1  = (x == width-1 ? D_U1(y, x - 1) : D_U1(y, x + 1));
        real_t west1  = (x == 0 ? D_U1(y, x + 1) : D_U1(y, x - 1));
        real_t north1 = D_U1(y - 1, x);
        real_t south1 = D_U1(y + 1, x);

        // handle the standard inner case with 5 point star stencil
        D_U2(y, x) = 2.0f * mid1 - D_U0(y, x) + (c * c) * (-4.0f * mid1 + east1 + south1 + north1 + west1);
    }
}
extern "C" __host__ void gpu_main_kernel(kernel_type kernel){
    if (kernel == FULL){
        full_kernel<<<grid, block, 0, comp_stream>>>(c, d_u0, d_u1, d_u2, local_size[GPUANDCPU][X], local_size[GPUANDCPU][Y], ns);
    }
    else if (kernel == INNER){
        inner_kernel<<<inner_grid, block, 0, comp_stream>>>(c, d_u0, d_u1, d_u2, local_size[GPU][X], local_size[GPU][Y]);
    }
    else if (kernel == FULL_WITHOUT_INTRA_BORDER){
        gpu_share_without_intra_border_kernel<<<inner_grid, block, 0, comp_stream>>>(c, d_u0, d_u1, d_u2, local_size[GPU][X], local_size[GPU][Y], local_size[GPUANDCPU][Y]);
    }
}

extern "C" __host__ void gpu_border_kernel(border_kernel_type kernel){
    if (kernel == DEVICE_BORDER){
        outer_kernel<<<full_border_grid, full_border_block, 0, comp_stream>>>(c, d_u0, d_u1, d_u2, local_size[GPU][X], local_size[GPU][Y], local_size[GPUANDCPU][Y], ns);
    }
    else if (kernel == INTRA_BORDER){
        intra_border_kernel<<<single_border_grid, single_border_block, 0, comp_stream>>>(c, d_u0, d_u1, d_u2, local_size[GPU][X], local_size[GPU][Y]);
    }
    cudaEventRecord(border_done, comp_stream);
}

extern "C" __host__ void gpu_unpack_borders(border_kernel_type kernel){
    int width = local_size[GPU][X];
    if (kernel == DEVICE_BORDER){
        cudaMemcpyAsync(&D_U2(-1, 0), recv_borders, width * sizeof(real_t), cudaMemcpyHostToDevice, mem_stream);
        cudaMemcpyAsync(&D_U2(local_size[GPUANDCPU][Y], 0), &recv_borders[width], width * sizeof(real_t), cudaMemcpyHostToDevice, mem_stream);
    }
    else if (kernel == INTER_BORDER){
        cudaMemcpyAsync(&D_U2(-1, 0), recv_borders, width * sizeof(real_t), cudaMemcpyHostToDevice, mem_stream);
    }
}

extern "C" __host__ void gpu_pack_borders(border_kernel_type kernel){
    int width = local_size[GPU][X];
    if (kernel == DEVICE_BORDER){
        cudaMemcpyAsync(send_borders, &D_U2(0, 0), width * sizeof(real_t), cudaMemcpyDeviceToHost, mem_stream);
        cudaMemcpyAsync(&send_borders[width], &D_U2(local_size[GPUANDCPU][Y]-1, 0), width * sizeof(real_t), cudaMemcpyDeviceToHost, mem_stream);
    }
    else if(kernel == INTER_BORDER){
        cudaMemcpyAsync(send_borders, &D_U2(0, 0), width * sizeof(real_t), cudaMemcpyDeviceToHost, mem_stream);
    }
    cudaEventRecord(packing_done, mem_stream);
}

extern "C" __host__ void intra_border_exchange(){
    int_t width = local_size[GPU][X];
    if (local_size[CPU][Y] > 0){
        cudaMemcpyAsync(&D_U2(local_size[GPU][Y], 0), &H_U(2, local_size[GPU][Y], 0), width * sizeof(real_t), cudaMemcpyHostToDevice, mem_stream);
        cudaMemcpyAsync(&H_U(2, local_size[GPU][Y] - 1, 0), &D_U2(local_size[GPU][Y] - 1, 0), width * sizeof(real_t), cudaMemcpyDeviceToHost, mem_stream);
    }
}

extern "C" __host__ void fetch_gpu_grids(int_t timestep){
    cudaMemcpy(h_u[timestep], (timestep == 0 ? d_u0 : timestep == 1 ? d_u1: d_u2), local_size_bytes[GPU] - local_size[GPU][X] * sizeof(real_t), cudaMemcpyDeviceToHost);
}

extern "C" __host__ void init_cuda(){
    int n_devices;
    cudaGetDeviceCount(&n_devices);
    cudaSetDevice(rank % n_devices);

    cudaStreamCreate(&comp_stream);
    cudaStreamCreate(&mem_stream);

    cudaEventCreate(&start);
    cudaEventCreate(&stop);
    cudaEventCreate(&border_done);
    cudaEventCreate(&packing_done);
}

__global__ void initialize_state(real_t *d_u0, real_t *d_u1, real_t *d_u2, int_t local_origin_x, int_t local_origin_y, int_t width, int_t height, int_t global_size){
    int_t x = threadIdx.x + blockDim.x * blockIdx.x - 1;
    int_t y = threadIdx.y + blockDim.y * blockIdx.y - 1;
    if (x <= width && y <= height){
        D_U0(y, x) = -exp(
            -pow(local_origin_x + x - ((real_t)global_size / 2.0), 2.0) / (real_t)global_size 
            -pow(local_origin_y + y - ((real_t)global_size / 2.0), 2.0) / (real_t)global_size);
        D_U1(y, x) = D_U0(y, x);
        D_U2(y, x) = D_U0(y, x);
    }
}

extern "C" __host__ void inter_and_intra_calibration(real_t *rows_per_s){
    // we must calibrate the intra node balance first, as the combined cpu and gpu computational throughput
    // will determine the inter load balance
    cudaMalloc((void **)&d_u0, local_size_bytes[GPUANDCPU]);
    cudaMalloc((void **)&d_u1, local_size_bytes[GPUANDCPU]);
    if (0 != cudaMalloc((void **)&d_u2, local_size_bytes[GPUANDCPU])){
        printf("malloc failed!\n");
        exit(1);
    }

    h_u[0] = (real_t *) malloc (local_size_bytes[GPUANDCPU]);
    h_u[1] = (real_t *) malloc (local_size_bytes[GPUANDCPU]);
    h_u[2] = (real_t *) malloc (local_size_bytes[GPUANDCPU]);

    initialize_state_with_gpu(); // fills h_u and d_u
    dim3 block = dim3(32, 32);
    real_t gpu_ms, cpu_ms;

    if (intra_calibrate){ // calibrate the share of rows that should be computed on the gpu with a binary search
        dim3 grid = dim3(CEIL_DIV(local_size[GPU][X], 32), CEIL_DIV(local_size[GPU][Y], 32));
        struct binary_search_handler binsearch;
        binary_search_handler_setup(&binsearch, 1, local_size[GPUANDCPU][Y]-1, false);

        gpu_share_without_intra_border_kernel<<<grid, block, 0, comp_stream>>>(c, d_u0, d_u1, d_u2, local_size[GPU][X], local_size[GPU][Y], local_size[GPUANDCPU][Y]);
        
        while (!binsearch.converged){
            grid = dim3(CEIL_DIV(local_size[GPU][X], 32), CEIL_DIV(binsearch.mid, 32)); // update the share of the gpu before timing it

            // get gpu runtime of current config
            cudaEventRecord(start, comp_stream);
            gpu_share_without_intra_border_kernel<<<grid, block, 0, comp_stream>>>(c, d_u0, d_u1, d_u2, local_size[GPU][X], binsearch.mid, local_size[GPUANDCPU][Y]);
            cudaEventRecord(stop, comp_stream);
            cudaEventSynchronize(stop);
            cudaEventElapsedTime(&gpu_ms, start, stop);

            // get cpu runtime of current config
            cpu_ms = cpu_kernel_without_intra_border(h_u, local_size[GPUANDCPU][Y], binsearch.mid, local_size[GPU][X]);
            binary_search(&binsearch, cpu_ms, gpu_ms);
        }

        update_global_variables_with_new_gpu_share(binsearch.mid);
    }

    if (inter_calibrate){
        dim3 grid = dim3(CEIL_DIV(local_size[GPUANDCPU][X], 32), CEIL_DIV(local_size[GPUANDCPU][Y], 32));

        inner_kernel<<<grid, block, 0, comp_stream>>>(c, d_u0, d_u1, d_u2, local_size[GPUANDCPU][X], local_size[GPUANDCPU][Y]);
        cudaEventRecord(start, comp_stream);
        for (int i = 0; i < 10; i++)
            inner_kernel<<<grid, block, 0, comp_stream>>>(c, d_u0, d_u1, d_u2, local_size[GPUANDCPU][X], local_size[GPUANDCPU][Y]);
        cudaEventRecord(stop, comp_stream);
        cudaEventSynchronize(stop);
        cudaEventElapsedTime(&gpu_ms, start, stop);

        *rows_per_s = 0;

        if (gpu_work_share != 1.0){ // we should encorporate the computational capacity of the CPU into this node
            cpu_ms = 0.0;
            for (int i = 0; i < 10; i++){
                cpu_ms += cpu_kernel_without_intra_border(h_u, local_size[GPUANDCPU][Y], local_size[GPU][Y], local_size[CPU][X]);
            }
            *rows_per_s = (real_t) local_size[CPU][Y] / (cpu_ms / 1000.0);
        }

        *rows_per_s += (real_t) local_size[GPUANDCPU][Y] / (gpu_ms / 1000.0f);
        printf("rows_per_s: %f\n", *rows_per_s);
    }
    cudaFree(d_u0);
    cudaFree(d_u1);
    cudaFree(d_u2);
    free(h_u[0]);
    free(h_u[1]);
    free(h_u[2]);
}

extern "C" __host__ void cuda_sync(sync_type SYNC_TYPE){
    if (SYNC_TYPE == DEVICE){
        cudaDeviceSynchronize();
    }
    if (SYNC_TYPE == BORDER_DONE){
        cudaEventSynchronize(border_done);
    }
    if (SYNC_TYPE == PACKING_DONE){
        cudaEventSynchronize(packing_done);
    }
}

extern "C" void initialize_state_with_gpu(){
    dim3 fullblock = dim3(32, 32);
    dim3 fullgrid = dim3(CEIL_DIV(local_size[GPUANDCPU][X] + 2, 32), CEIL_DIV(local_size[GPUANDCPU][Y] + 2, 32));

    initialize_state<<<fullgrid, fullblock>>>(d_u0, d_u1, d_u2, local_origin[GPUANDCPU][X], local_origin[GPUANDCPU][Y], local_size[GPUANDCPU][X], local_size[GPUANDCPU][Y], global_size[X]);
    cudaMemcpy(h_u[0], d_u0, local_size_bytes[GPUANDCPU], cudaMemcpyDeviceToHost);
    cudaMemcpy(h_u[1], d_u1, local_size_bytes[GPUANDCPU], cudaMemcpyDeviceToHost);
    cudaMemcpy(h_u[2], d_u2, local_size_bytes[GPUANDCPU], cudaMemcpyDeviceToHost);
}

extern "C" __host__ void set_cuda_grids(){
    block = dim3(32, 32);
    grid = dim3(CEIL_DIV(local_size[GPU][X], 32), CEIL_DIV(local_size[GPU][Y], 32));

    full_border_block = dim3(MIN(1024, mpi_circumference));
    full_border_grid = dim3(CEIL_DIV(local_size[GPU][X]*2, full_border_block.x));
    single_border_block = dim3(MIN(1024, local_size[GPU][X]));
    single_border_grid = dim3(CEIL_DIV(local_size[GPU][X], single_border_block.x));

    inner_grid = dim3(CEIL_DIV(local_size[GPU][X], 32), CEIL_DIV(local_size[GPU][Y] - 1, 32));
}

extern "C" __host__ void initialize_cuda_mem(){
    // allocate pinned memory for faster memory transfers
    if (0 != cudaMallocHost((void **)&send_borders, mpi_circumference * sizeof(real_t))){
        printf("mallochost failed!\n");
        exit(1);
    }
    if (0 != cudaMallocHost((void **)&recv_borders, mpi_circumference * sizeof(real_t))){
        printf("mallochost failed!\n");
        exit(1);
    }

    cudaMalloc((void **)&d_u0, local_size_bytes[GPUANDCPU]);
    cudaMalloc((void **)&d_u1, local_size_bytes[GPUANDCPU]);
    if (0 != cudaMalloc((void **)&d_u2, local_size_bytes[GPUANDCPU])){
        printf("malloc failed!\n");
        exit(1);
    }
}


extern "C" void get_gputime(real_t *gputime){
    cudaEventSynchronize(stop);
    cudaEventElapsedTime(gputime, start, stop);
}

extern "C" __host__ void cuda_terminate(){
    cudaFree(d_u0);
    cudaFree(d_u1);
    cudaFree(d_u2);
    cudaFreeHost(recv_borders);
    cudaFreeHost(send_borders);
}
