#include <mpi.h>

// TEMP
#define _GNU_SOURCE
#include <sched.h>

#include "heterogeneous_FDM_config.h"

/*
compilation example:
mpicc -c heterogeneous_FDM_solver.c -lm -O3 -Wall -fopenmp -DITERATIONS=40000 -DANIMATE=true -DSIZE=128 -DSNAP_FREQ=200 -DGPUSHARE=1.0 -DINTER_CALIBRATE=false -o heterogeneous_FDM_solver.o
mpicc -c heterogeneous_FDM_CPU_kernels.c -lm -O3 -Wall -fopenmp -DITERATIONS=40000 -DANIMATE=true -DSIZE=128 -DSNAP_FREQ=200 -DGPUSHARE=1.0 -DINTER_CALIBRATE=false -o cpu_kernels.o
mpicc -c heterogeneous_FDM_config.c -lm -O3 -Wall -fopenmp -DITERATIONS=40000 -DANIMATE=true -DSIZE=128 -DSNAP_FREQ=200 -DGPUSHARE=1.0 -DINTER_CALIBRATE=false -o heterogeneous_FDM_config.o
nvcc -c heterogeneous_FDM_GPU_kernels.cu -lstdc++ -lcudart -DITERATIONS=40000 -DANIMATE=true -DSIZE=128 -DSNAP_FREQ=200 -DGPUSHARE=1.0 -DINTER_CALIBRATE=false -o gpu_kernels.o
mpicc heterogeneous_FDM_solver.o heterogeneous_FDM_config.o gpu_kernels.o cpu_kernels.o -lstdc++ -lcudart -lm -O3 -Wall -fopenmp -o heterogeneous_FDM_solver
*/

MPI_Datatype subdomain, domain;
MPI_Request send_recv_reqs_nesw[4]; // a request array in the following order: [NorthReq,SouthReq]

int main(int argc, char **argv)
{
    double t1, t2;
    // double outer_comp = 0.0, comp = 0.0, comm = 0.0;

    setbuf(stdout, NULL);

    init_mpi();
    init_cuda();
    inter_and_intra_calibration(&row_per_s);                         // computes gpu_work_share, and also ideal subgrid height
    distribute_local_shares();                                       // distributes gpuheight and origin_y into local_size, local_origin
    update_global_variables_with_new_subdomain_size(gpu_work_share); // update the local dimensions

    // Initialize memory for the local grid, create MPI types, initialize memory on CUDA capable device
    initialize_global_mem();
    initialize_MPI_types();
    initialize_cuda_mem();
    set_cuda_grids();
    initialize_state_with_gpu(); // use the GPU to set the values in the matrix to save time

    const bool ONLY_GPU = gpu_work_share == 1.00;
    const bool NO_MPI = size == 1;

    cuda_sync(DEVICE); // synchronize before timing
    MPI_Barrier(MPI_COMM_WORLD);
    t1 = MPI_Wtime();

    // printf("P%d={loc_size_gpu:[%d,%d], loc_size_cpu:[%d,%d], loc_size:[%d,%d]}, byte_loc:[%d,%d,%d], ns: %d,%d\n", rank, local_size[GPU][X], local_size[GPU][Y], local_size[CPU][X], local_size[CPU][Y], local_size[GPUANDCPU][X], local_size[GPUANDCPU][Y], local_size_bytes[GPU], local_size_bytes[CPU], local_size_bytes[GPUANDCPU], north, south);
    // real_t gpu_border = 0, cpu_border=0, gpu_inner = 0, intra_ex = 0, cpu_inner = 0; 
    // struct timeval a, b;
    for (int_t iter = 0; iter < max_iter + 1; iter++)
    {
        if (animate && iter % snap_freq == 0)
        {
            fetch_gpu_grids(0);
            cuda_sync(DEVICE);
            save_state(iter);
            if (rank == 0)
            {
                printf("%ld / %ld\n", iter, max_iter);
            }
        }

        if (ONLY_GPU)
        {
            if (NO_MPI)
            { // only gpu and no mpi means it suffices to queue kernels
                gpu_main_kernel(FULL);
            }
            else
            { // Only GPU, multiple processes means we must also perform a border exchange
                gpu_border_kernel(DEVICE_BORDER); 
                gpu_main_kernel(INNER); 
                cuda_sync(BORDER_DONE);
                gpu_pack_borders(DEVICE_BORDER); 

                cuda_sync(PACKING_DONE);
                inter_border_exchange(ONLY_GPU); // initate border exchange between CPUs
                MPI_Waitall(4, send_recv_reqs_nesw, MPI_STATUSES_IGNORE); // wait for border exchange
                
                gpu_unpack_borders(DEVICE_BORDER); 
            }
        }
        else
        {
            if (NO_MPI)
            { // local GPU and CPU, needs internal border exchange
                gpu_border_kernel(INTRA_BORDER);
                gpu_main_kernel(FULL_WITHOUT_INTRA_BORDER);
                cpu_intra_border_kernel();
                
                cuda_sync(BORDER_DONE);
                intra_border_exchange();

                cpu_kernel_without_intra_border(h_u, local_size[GPUANDCPU][Y], local_size[GPU][Y], local_size[CPU][X]);
            }
            else
            { // local GPU and CPU with multiple ranks, internal and external border exchange
                gpu_border_kernel(DEVICE_BORDER);
                gpu_main_kernel(INNER);
                cpu_border_kernel();
                cuda_sync(BORDER_DONE); 
                gpu_pack_borders(INTER_BORDER);
                
                cuda_sync(PACKING_DONE);
                inter_border_exchange(ONLY_GPU);
                intra_border_exchange();

                cpu_inner_kernel(h_u);
                MPI_Waitall(4, send_recv_reqs_nesw, MPI_STATUSES_IGNORE); // wait for border exchange

                gpu_unpack_borders(INTER_BORDER);
            }
        }
        cuda_sync(DEVICE);
        rotate_buffers();
    }

    cuda_sync(DEVICE);                 // make sure CUDA computations are complete
    MPI_Barrier(MPI_COMM_WORLD); // make sure all processes are done

    t2 = MPI_Wtime();

    if (rank == 0)
    {
        double tot = (t2 - t1);
        printf("\n%lf\n", tot);
        // printf("SIZE: %d\nGPUOUTER: %lf\nCPU_OUTER: %lf\nGPUINNER: %lf\nINTRA_EX: %lf\nCPU_INNER: %lf\n", sidelength, gpu_border, cpu_border, gpu_inner, intra_ex, cpu_inner);
        // printf("[%ldx%ld] %d processes, total_runtime: %lfs\n", sidelength, sidelength, size, tot);
        // printf("%.3fit/s, %.3fpt/s, %fs/it\n\n", max_iter/tot, sidelength*sidelength*max_iter/tot, tot/max_iter);
    }

    free(h_u[0]);
    free(h_u[1]);
    free(h_u[2]);
    MPI_Type_free(&domain);
    MPI_Type_free(&subdomain);
    MPI_Finalize();
    cuda_terminate();
    exit(EXIT_SUCCESS);
}

void rotate_buffers()
{
    real_t *tmp = d_u0;
    d_u0 = d_u1;
    d_u1 = d_u2;
    d_u2 = tmp;

    tmp = h_u[0];
    h_u[0] = h_u[1];
    h_u[1] = h_u[2];
    h_u[2] = tmp;
}

void init_mpi()
{
    MPI_Init(NULL, NULL);
    MPI_Comm_size(MPI_COMM_WORLD, &size);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank); // fetch this process' rank

    north = (rank == 0 ? MPI_PROC_NULL : rank - 1);
    south = (rank == size - 1 ? MPI_PROC_NULL : rank + 1);

    ns = 0; // init the bitmask telling whether a process in on an edge
    ns |= (north < 0) << 0;
    ns |= (south < 0) << 1;

    global_size[X] = global_size[Y] = sidelength;
    local_size[GPUANDCPU][Y] = global_size[Y] / size; // add condition here giving extra rows to last rank if ranks to not divide sidelength
    local_size[GPU][Y] = (int)local_size[GPUANDCPU][Y] * gpu_work_share;
    local_size[CPU][Y] = local_size[GPUANDCPU][Y] - local_size[GPU][Y];
    local_size[GPUANDCPU][X] = local_size[GPU][X] = local_size[CPU][X] = global_size[X];
    local_origin[GPUANDCPU][X] = local_origin[GPU][X] = local_origin[CPU][X] = local_origin[GPUANDCPU][Y] = local_origin[GPU][Y] = 0;
    local_origin[CPU][Y] = local_size[GPU][Y];
    mpi_circumference = local_size[GPUANDCPU][X] * 2;
    local_size_bytes[GPUANDCPU] = (local_size[GPUANDCPU][X] + 2) * (local_size[GPUANDCPU][Y] + 2) * sizeof(real_t);
}

void initialize_MPI_types(void)
{
    MPI_Type_create_subarray(
        2,
        (int[2]){local_size[GPUANDCPU][Y] + 2, local_size[GPUANDCPU][X] + 2},
        local_size[GPUANDCPU],
        (int[2]){1, 1},
        MPI_ORDER_C, MPI_FLOAT, &subdomain);
    MPI_Type_create_subarray(
        2, global_size, local_size[GPUANDCPU], local_origin[GPUANDCPU],
        MPI_ORDER_C, MPI_FLOAT, &domain);
    MPI_Type_commit(&subdomain);
    MPI_Type_commit(&domain);
}

// Have all processes estimate their local computational power
// then give a proportional number of global rows to each process
// based on their computational throughput
void distribute_local_shares()
{
    real_t *all_rows_per_s;
    int *all_local_rows;
    int *orig_y;
    // didnt really want to allocate this everywhere but got compiler warnings
    all_rows_per_s = malloc(size * sizeof(real_t));
    all_local_rows = malloc(size * sizeof(int_t));
    orig_y = malloc(size * sizeof(int_t));
    orig_y[0] = 0;
    MPI_Gather(&row_per_s, 1, MPI_FLOAT, all_rows_per_s, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);

    if (rank == 0)
    {
        real_t sum = 0;
        int_t rows_left = global_size[Y];
        for (int i = 0; i < size; i++)
        {
            sum += all_rows_per_s[i];
        }
        for (int i = 0; i < size; i++)
        {
            if (i == size - 1)
            {
                all_local_rows[i] = rows_left;
            }
            else
            {
                all_local_rows[i] = (int_t)((real_t)global_size[Y] * (all_rows_per_s[i] / sum));
                rows_left -= all_local_rows[i];
            }

            if (i < size - 1)
            {
                orig_y[i + 1] = orig_y[i] + all_local_rows[i];
            }
        }
    }

    MPI_Scatter(all_local_rows, 1, MPI_INT, &local_size[GPUANDCPU][Y], 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Scatter(orig_y, 1, MPI_INT, &local_origin[GPUANDCPU][Y], 1, MPI_INT, 0, MPI_COMM_WORLD);
    // printf("\np[%d]=%d rows\n", rank, local_size[Y]);
    free(all_local_rows);
    free(all_rows_per_s);
}

// only used during the binary search to make sure the new correct local sizes are used
void update_global_variables_with_new_gpu_share(int_t gpu_y_dim)
{
    local_size[GPU][Y] = gpu_y_dim;
    local_size[CPU][Y] = local_size[GPUANDCPU][Y] - local_size[GPU][Y];
    local_origin[CPU][Y] = local_origin[GPU][Y] + local_size[GPU][Y];
}
void update_global_variables_with_new_subdomain_size(real_t gpu_work_share)
{
    local_size[GPU][Y] = (int)((double)local_size[GPUANDCPU][Y] * gpu_work_share);
    local_size[CPU][Y] = local_size[GPUANDCPU][Y] - local_size[GPU][Y];
    local_origin[GPU][Y] = local_origin[GPUANDCPU][Y]; // origin[GPUANDCPU][Y] was set by the gpu share distribution
    local_origin[CPU][Y] = local_origin[GPU][Y] + local_size[GPU][Y];
    local_size_bytes[GPUANDCPU] = (local_size[GPUANDCPU][Y] + 2) * (local_size[GPUANDCPU][X] + 2) * sizeof(real_t);
    local_size_bytes[GPU] = (local_size[GPU][Y] + 2) * (local_size[GPU][X] + 2) * sizeof(real_t);
    local_size_bytes[CPU] = (local_size[CPU][Y] + 2) * (local_size[CPU][X] + 2) * sizeof(real_t);
}

void initialize_global_mem()
{
    h_u[0] = malloc(local_size_bytes[GPUANDCPU]);
    h_u[1] = malloc(local_size_bytes[GPUANDCPU]);
    h_u[2] = malloc(local_size_bytes[GPUANDCPU]);
}

void inter_border_exchange(bool ONLY_GPU)
{
    int_t width = local_size[GPUANDCPU][X];
    int_t north_idx = 0, south_idx = width;

    float *send_north = &send_borders[north_idx];
    float *recv_north = &recv_borders[north_idx];

    // If were using the CPU, we dont need to waste time packing/unpacking the results if MPI simply sends them to the right buffer
    float *send_south = (ONLY_GPU ? &send_borders[south_idx] : &H_U(2, local_size[GPUANDCPU][Y] - 1, 0));
    float *recv_south = (ONLY_GPU ? &recv_borders[south_idx] : &H_U(2, local_size[GPUANDCPU][Y], 0));

    // Send north receive south
    MPI_Isend(send_north, width, MPI_FLOAT, north, 0, MPI_COMM_WORLD, &send_recv_reqs_nesw[0]);
    MPI_Irecv(recv_south, width, MPI_FLOAT, south, 0, MPI_COMM_WORLD, &send_recv_reqs_nesw[1]);

    // Send south receive north
    MPI_Isend(send_south, width, MPI_FLOAT, south, 1, MPI_COMM_WORLD, &send_recv_reqs_nesw[2]);
    MPI_Irecv(recv_north, width, MPI_FLOAT, north, 1, MPI_COMM_WORLD, &send_recv_reqs_nesw[3]);
}

void save_state(int_t iter)
{
    char filename[256];
    memset(filename, 0, 256);
    sprintf(filename, "data/%.5d.dat", (int)(iter / snap_freq)); // add data/

    MPI_File out;
    MPI_File_open(
        MPI_COMM_WORLD, filename, MPI_MODE_CREATE | MPI_MODE_WRONLY, MPI_INFO_NULL, &out);
    MPI_File_set_view(
        out, 0, MPI_FLOAT, domain, "native", MPI_INFO_NULL);
    MPI_File_write_all(
        out, &H_U(0, -1, -1), 1, subdomain, MPI_STATUS_IGNORE);
    MPI_File_close(&out);
}