#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/time.h>
#include <string.h>
#include <mpi.h>

/*
ping pong test between all pairs of MPI ranks
also performs linear regression on the results to retrieve the latency and bandwidth
*/

int rank, size, n_tests=100000, max_kb=100;
char *buf;

double time_diff_in_ms(struct timeval start, struct timeval end) {
    long seconds = end.tv_sec - start.tv_sec;
    long microseconds = end.tv_usec - start.tv_usec;
    double milliseconds = (seconds * 1000) + (microseconds / 1000.0);
    return milliseconds;
}

void linearRegression(double x[], double y[], int n, double *slope, double *intercept) {
    double sumX = 0.0, sumY = 0.0, sumXY = 0.0, sumXX = 0.0;
    
    // Calculate the sums
    for (int i = 0; i < n; i++) {
        sumX += x[i];
        sumY += y[i];
        sumXY += x[i] * y[i];
        sumXX += x[i] * x[i];
    }

    // Calculate the slope (beta1)
    double numerator = (n * sumXY) - (sumX * sumY);
    double denominator = (n * sumXX) - (sumX * sumX);
    *slope = numerator / denominator;
    
    // Calculate the intercept (beta0)
    *intercept = (sumY - (*slope * sumX)) / n;
}
// get time of sending messages between two mpi ranks
double pingpong(int64_t num_of_bytes, int from, int to){
    double t1=0.0, t2=0.0;
    struct timeval cpuA, cpuB;
    MPI_Barrier(MPI_COMM_WORLD);
    gettimeofday(&cpuA, NULL);
    for (int i = 0; i < n_tests; i++){
        if (rank == from){
            MPI_Ssend(buf, num_of_bytes, MPI_BYTE, to, 0, MPI_COMM_WORLD); 
            MPI_Recv(buf, num_of_bytes, MPI_BYTE, to, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        }else if (rank == to){
            MPI_Recv(buf, num_of_bytes, MPI_BYTE, from, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            MPI_Ssend(buf, num_of_bytes, MPI_BYTE, from, 0, MPI_COMM_WORLD);
        }
    }
    gettimeofday(&cpuB, NULL);
    
    return (time_diff_in_ms(cpuA, cpuB))/(n_tests*2.0);
}

int main(int argc, char *argv[], char * envp[]){

    setbuf(stdout, NULL);

    MPI_Init(NULL, NULL);
    MPI_Comm_rank ( MPI_COMM_WORLD, &rank );
    MPI_Comm_size ( MPI_COMM_WORLD, &size );

    buf = (char*)calloc(1, sizeof(char)*max_kb*1024);
    if (buf == NULL){
        printf("unable to allocate enough mem\n");
        exit(1);
    }

    char* env_var = "SLURMD_NODENAME=";
    for (int i = 0; envp[i] != NULL; i++)
    {   
        if (strstr(envp[i], env_var) != NULL) printf("\nrank=%d, node=%s\n", rank, envp[i]);
    }

    double* X = malloc((max_kb+1)*sizeof(double));
    double* Y = malloc((max_kb+1)*sizeof(double));
    double slope, intercept;
    double *slopes = malloc(sizeof(double)*size*size);
    double *intercepts = malloc(sizeof(double)*size*size);

    for (int i = 0; i < size*size; i++){
        slopes[i] = 0;
        intercepts[i] = 0;
    }

    for (int i = 0; i <= max_kb; i++) X[i] = (double) i;

    // measure for every pair of MPI ranks, assume symmetry
    for (int A = 0; A < size; A++){
        for (int B = 0; B < size; B++){
            if (B > A){ 
                for (int n_kb = 0; n_kb <= max_kb; n_kb++){
                    Y[n_kb] = pingpong(1024*n_kb, A, B);
                    // if (rank == A){
                    //     printf("%lf\n", Y[n_kb]);
                    // }
                }
                if (rank == B){
                    linearRegression(X, Y, max_kb+1, &slope, &intercept);
                }

                // send results to rank 0 for output in correct order
                if (rank == B || rank == 0){
                    int recipient = (rank == B ? 0 : MPI_PROC_NULL);
                    int sender    = (rank == 0 ? B : MPI_PROC_NULL);
                    MPI_Sendrecv(&slope, 1, MPI_DOUBLE, recipient, 0, &slopes[A*size+B], 1, MPI_DOUBLE, sender, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
                    MPI_Sendrecv(&intercept, 1, MPI_DOUBLE, recipient, 0, &intercepts[A*size+B], 1, MPI_DOUBLE, sender, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
                }
            }
        }
    }

    printf("\n");
    if (rank == 0){
        for (int i = 0; i < size; i++){
            for (int j = 0; j < size; j++){
                printf("%lf ", slopes[i*size + j]);
            }
            printf("\n");
        }
    }
    printf("\n");
    if (rank == 0){
        for (int i = 0; i < size; i++){
            for (int j = 0; j < size; j++){
                printf("%lf ", intercepts[i*size + j]);
            }
            printf("\n");
        }
    }

    MPI_Finalize();
    return 0;
}